#!/bin/python
#-*- coding: utf-8 -*-

# Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS

# This software is published by the license of CPU-OS License, you can use and 
# distibuted this software under this License. See CPU-OS Liense for more detail.

# You should have receive a copy of CPU-OS License. If not, please contact us 
# by email <support_os@cpu-os.ac.cn>  

import apt_pkg
import os
import tempfile
import urllib2
import logging
import traceback
import time
import sys
import gnupg
import zipfile

class Upgrade_Meta_Info_Obj(object):
    def __init__(self, cur, cur_version, cur_version_id, cur_time, dist, dist_version, dist_version_id, dist_time, supported):
        self.cur = cur
        self.cur_version = cur_version
        self.cur_version_id = cur_version_id
        self.cur_time = cur_time
        self.dist = dist
        self.dist_version = dist_version
        self.dist_version_id = dist_version_id
        self.dist_time = dist_time
        try:
            self.supported = bool(int(supported))
        except Exception as e:
            logging.error( "init Upgrade_Meta_Info_Obj supported Exception:" + str(e))
            self.supported = False

        self.description = None
        self.releaseFileURI = None
        self.releaseNotesURI = None
        self.upgradeTool = None
        self.upgradeToolSig = None
        # the server may report that the upgrade is broken currently
        self.upgrade_broken = None

#从/etc/os-release中获取系统版本信息.
'''
def get_current_version_with_time():
    dist_file = open("/etc/os-release")
    dist_info = ""
    for line in dist_file.read().split("\n"):
        if line.startswith("VERSION="):
            dist_info = line.split("\"")[1]
    return dist_info
'''

def get_current_version_with_time():
    logging.debug("get_current_version_with_time")
    dist_file = open("/etc/os-release", "r")
    for line in dist_file.read().split("\n"):
        if line.startswith("VERSION_ID"):
            version_id = line.split("=")[1]
    
    cpuinfo_file = os.popen("grep 'model name' /proc/cpuinfo", "r", -1)
    model_name = cpuinfo_file.read().split("\n")[0].split(": ")[1]
    
    if model_name.lower().find("hygon") != -1:
        cpu_name = "Hygon"
    if model_name.lower().find("loongson") != -1:
        cpu_name = "Loongson"
    if model_name.lower().find("phytium") != -1:
        cpu_name = "Phytium"
    if model_name.lower().find("sw") != -1:
        cpu_name = "Sunway"
    
    dist_info = cpu_name + ";" + version_id
    logging.debug("dist_info= %s" %(dist_info))
    
    return dist_info

class MetaReleaseCore(object):

    def __init__(self,
                 METARELEASE_URI,
                 cache=None):
        self.METARELEASE_URI = METARELEASE_URI
        self.current_os_obj = None
        self._buildMetaReleaseFile()

    #将meta-release文件及相关的文件的全路径存放到成员变量中.
    def _buildMetaReleaseFile(self):
        # build the metarelease_file name
        self.METARELEASE_FILE_DIR = "/usr/share/cdos-update/upgrade/"
        if not os.path.isdir(self.METARELEASE_FILE_DIR):
            os.makedirs(self.METARELEASE_FILE_DIR, mode=0o777)
        # meta-release
        self.METARELEASE_FILE = os.path.join(
            "/usr/share/cdos-update/upgrade/",
            os.path.basename(self.METARELEASE_URI))
        # gpg for meta-release
        self.METARELEASE_SIGNATURE_URI = self.METARELEASE_URI + ".gpg"
        self.METARELEASE_SIGNATURE_FILE = os.path.join(
            "/usr/share/cdos-update/upgrade/",
            os.path.basename(self.METARELEASE_SIGNATURE_URI))

    def download_meta_file(self):
        # self._buildMetaReleaseFile()
        self.download_file( self.METARELEASE_URI, self.METARELEASE_FILE)

    def download_meta_singature(self):
        self.download_file(self.METARELEASE_SIGNATURE_URI, self.METARELEASE_SIGNATURE_FILE)

    def download_meta_file_with_verify(self):
        self.download_meta_file()
        self.download_meta_singature()
        # ToDo
        if self.__gpg_check_patch(self.METARELEASE_SIGNATURE_FILE, self.METARELEASE_FILE) == True:
            return True
        else:
            return False

    def get_meta_file_exist(self, gpg_check_flag=False):
        if gpg_check_flag == False:
            return os.path.exists(self.METARELEASE_FILE)
        else:
            if os.path.exists(self.METARELEASE_FILE) and os.path.exists(self.METARELEASE_SIGNATURE_FILE):
                return self.__gpg_check_patch(self.METARELEASE_SIGNATURE_FILE, self.METARELEASE_FILE)
            else:
                return False

    def get_upgrade_tool_path(self):
        upgrade_tool_url = self.current_os_obj.upgradeTool
        upgrade_tool_path = os.path.join(
            "/usr/share/cdos-update/upgrade/",
            os.path.basename(upgrade_tool_url))
        return upgrade_tool_path

    #解压升级工具包.
    def extract_upgrade_tools(self):
        upgrade_tool_url = self.current_os_obj.upgradeTool
        upgrade_tool_path = os.path.join(
            "/usr/share/cdos-update/upgrade/",
            os.path.basename(upgrade_tool_url))
        upgrade_tool_dir = upgrade_tool_path[:-4]
        self.__unzip_tools( upgrade_tool_path, upgrade_tool_dir)
        return upgrade_tool_dir

    def prepare_upgrade_tool(self, upgrade_tool_dir=None):
        if upgrade_tool_dir is None:
            upgrade_tool_url = self.current_os_obj.upgradeTool
            upgrade_tool_path = os.path.join(
                "/usr/share/cdos-update/upgrade/",
                os.path.basename(upgrade_tool_url))
            upgrade_tool_dir = upgrade_tool_path[:-4]

        if os.path.exists("/usr/share/cdos-update/upgrade/tools"):
            if os.path.isdir("/usr/share/cdos-update/upgrade/tools"):
                os.rmdir("/usr/share/cdos-update/upgrade/tools")
            else:
                os.remove("/usr/share/cdos-update/upgrade/tools")
        os.rename( upgrade_tool_dir, "/usr/share/cdos-update/upgrade/tools")

    #解压zip包
    def __unzip_tools(self, tool_path, tool_dir):
        zfile = zipfile.ZipFile(tool_path)
        zfile.extractall(tool_dir)
        zfile.close()

    #下载文件，下载url以及存放目录.
    def download_file(self, url, dist_file):
        dist_store_file = dist_file
        downloading_file = tempfile.mktemp(os.path.basename(dist_file))
        try:
            if os.path.exists(dist_store_file):
                os.remove(dist_store_file)
            if os.path.exists(downloading_file):
                os.remove(downloading_file)

            u = urllib2.urlopen(url, timeout=10)
            f = open(downloading_file, 'wb')
            meta = u.info()
            file_size = int(meta.getheaders("Content-Length")[0])
            logging.debug("Downloading: %s Bytes: %s" % (dist_store_file, file_size))

            file_size_dl = 0
            block_sz = 8192
            while True:
                buffer = u.read(block_sz)
                if not buffer:
                    break

                file_size_dl += len(buffer)
                f.write(buffer)
                status = r"%10d  [%3.2f%%]" % (file_size_dl, file_size_dl * 100. / file_size)
                status = status + chr(8)*(len(status)+1)

            f.close()
            os.rename(downloading_file, dist_store_file)
            # time.sleep(1)
        except urllib2.HTTPError as e1:
            logging.error('HTTPError exception %s:  %s' % (e1, traceback.format_exc()))
            raise urllib2.HTTPError(e1)
        except urllib2.URLError as e2:
            logging.error('URLError exception %s:  %s' % (e2, traceback.format_exc()))
            raise urllib2.URLError(e2)
        except Exception as e3:
            logging.error('generic exception %s:  %s' % (e3, traceback.format_exc()))
            raise Exception(e3)

    def get_meta_info(self, file_name=""):
        file_name = self.METARELEASE_FILE
        file_handler = open(file_name)
        meta_infos = self.get_meta_info_from_handler(file_handler)
        file_handler.close()
        return meta_infos

    #获取当前release版本的metainfo
    def get_current_os_upgrade_info(self, use_new_meta_data=False, check_flag=True, current_os=None):
        try:
            self.current_os_obj = None
            # if need download or meta-release can not be trusted
            if ( use_new_meta_data == True) or ( self.get_meta_file_exist( gpg_check_flag=check_flag) == False):
                if self.download_meta_file_with_verify() == False:
                    return None
            meta_infos = self.get_meta_info(file_name=self.METARELEASE_FILE)
            if current_os is None:
                current_os_version = get_current_version_with_time()
            for meta_info in meta_infos:
                if meta_info.cur_version == current_os_version:
                    self.current_os_obj = meta_info
                    return meta_info
            return None
        except Exception as e:
            return None

    #gpg验证升级工具.
    def  __gpg_check_patch(self, signed_file, verify_file):
        keypath = sys.path[0] + '/keyring'
        gpg = gnupg.GPG( gnupghome = keypath)
        verify = gpg.verify_file( open(signed_file, "rb"), verify_file)
        #对gpg结果进行判断
        if 'valid' in verify.__dict__ :
            if verify.valid :
                return True
            else:
                return False
        else:
            return False

    def download_upgrade_tools(self):
        if self.current_os_obj is None:
            self.get_current_os_upgrade_info()
        if self.current_os_obj is None:
            raise Exception("Not current os info")

        try:
            upgrade_tool_url = self.current_os_obj.upgradeTool
            upgrade_tool_sig_url = self.current_os_obj.upgradeToolSig
            upgrade_tool_path = os.path.join(
                "/usr/share/cdos-update/upgrade/",
                os.path.basename(upgrade_tool_url))
            upgrade_tool_sig_path = os.path.join(
                "/usr/share/cdos-update/upgrade/",
                os.path.basename(upgrade_tool_sig_url))

            self.download_file( upgrade_tool_url, upgrade_tool_path)

            self.download_file( upgrade_tool_sig_url, upgrade_tool_sig_path)

            if self.__gpg_check_patch( upgrade_tool_sig_path, upgrade_tool_path) == True:
                return True, upgrade_tool_path
            return False, ""

        except Exception as e:
            return False, ""

    # return status upgrade_tool_path current_os_obj，下载系统升级工具包.
    def download_upgrade_tools2(self):
        if self.current_os_obj is None:
            self.get_current_os_upgrade_info()
        if self.current_os_obj is None:
            raise Exception("Not current os info")

        try:
            upgrade_tool_url = self.current_os_obj.upgradeTool
            upgrade_tool_sig_url = self.current_os_obj.upgradeToolSig
            upgrade_tool_path = os.path.join(
                "/usr/share/cdos-update/upgrade/",
                os.path.basename(upgrade_tool_url))
            upgrade_tool_sig_path = os.path.join(
                "/usr/share/cdos-update/upgrade/",
                os.path.basename(upgrade_tool_sig_url))

            self.download_file( upgrade_tool_url, upgrade_tool_path)

            self.download_file( upgrade_tool_sig_url, upgrade_tool_sig_path)

            if self.__gpg_check_patch( upgrade_tool_sig_path, upgrade_tool_path) == True:
                return True, upgrade_tool_path, self.current_os_obj
            return False, "", self.current_os_obj

        except Exception as e:
            return False, "", self.current_os_obj

    #判断升级工具是否已经下载到本地目录，同时返回本地目录.
    def get_upgrade_tool_exist(self):
        if self.current_os_obj is None:
            self.get_current_os_upgrade_info()
        if self.current_os_obj is None:
            raise Exception("Not current os info")

        try:
            #获取升级工具及其签名下载的url，用于获取升级工具的名称.
            upgrade_tool_url = self.current_os_obj.upgradeTool
            upgrade_tool_sig_url = self.current_os_obj.upgradeToolSig
            #升级工具本地存放目录.
            upgrade_tool_path = os.path.join(
                "/usr/share/cdos-update/upgrade/",
                os.path.basename(upgrade_tool_url))
            upgrade_tool_sig_path = os.path.join(
                "/usr/share/cdos-update/upgrade/",
                os.path.basename(upgrade_tool_sig_url))

            if os.path.exists(upgrade_tool_path) and os.path.exists(upgrade_tool_sig_path) and self.__gpg_check_patch( upgrade_tool_sig_path, upgrade_tool_path) == True:
                return True, upgrade_tool_path
            return False, ""

        except Exception as e:
            return False, ""

    #meta-release解析，返回一个list，成员Upgrade_Meta_Info_Obj.
    def get_meta_info_from_handler(self, file_handler):
        file_tag = apt_pkg.TagFile(file_handler)
        meta_infos = []
        try:
            while file_tag.step():
                for required_key in ("Current", "Current_Version", "Current_Version_ID", "Current_Time", "Dist", "Dist_Version", "Dist_Version_ID", "Dist_Time", "Supported"):
                    if required_key not in file_tag.section:
                        raise Exception("Required key '%s' missing" % required_key)
                cur = file_tag.section["Current"]
                cur_version = file_tag.section["Current_Version"]
                cur_version_id = file_tag.section["Current_Version_ID"]
                cur_time = file_tag.section["Current_Time"]
                dist = file_tag.section["Dist"]
                dist_version = file_tag.section["Dist_Version"]
                dist_version_id = file_tag.section["Dist_Version_ID"]
                dist_time = file_tag.section["Dist_Time"]
                supported = file_tag.section["Supported"]
                meta_info = Upgrade_Meta_Info_Obj(cur, cur_version, cur_version_id, cur_time, dist, dist_version, dist_version_id, dist_time, supported)
                if "Description" in file_tag.section:
                    meta_info.description = file_tag.section["Description"]

                if "Release-File" in file_tag.section:
                    meta_info.releaseFileURI = file_tag.section["Release-File"]

                if "ReleaseNotes" in file_tag.section:
                    meta_info.releaseNotesURI = file_tag.section["ReleaseNotes"]

                if "UpgradeTool" in file_tag.section:
                    meta_info.upgradeTool = file_tag.section["UpgradeTool"]

                if "UpgradeToolSignature" in file_tag.section:
                    meta_info.upgradeToolSig = file_tag.section["UpgradeToolSignature"]

                if "UpgradeBroken" in file_tag.section:
                    meta_info.upgrade_broken = file_tag.section["UpgradeBroken"]

                meta_infos.append(meta_info)
        except Exception as e:
            raise Exception("Unable to parse %s" % self.METARELEASE_URI)
        return meta_infos



if __name__ == "__main__":
    try:
        print get_current_version_with_time()

        meta_obj = MetaReleaseCore("http://repos.01os.org/data/cdos-update/cdos-upgrade-new-server/meta-release")
        print meta_obj.download_meta_file_with_verify()

        meta_info = meta_obj.get_current_os_upgrade_info()
        if meta_info is not None:
            print meta_info.cur_version
            print meta_info.dist_version
        else:
            print None

        print meta_obj.download_upgrade_tools()
        print meta_obj.download_upgrade_tools2()

    except Exception as e:
        print e
